@@ -1,5 +1,3 @@
-#   Copyright (c) 2007-2008 Open Source Applications Foundation
-#
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
@@ -48,18 +46,43 @@ else:
 
 
 class JavaError(Exception):
+
     def getJavaException(self):
         return self.args[0]
 
+    def __str__(self):
+        writer = StringWriter()
+        self.getJavaException().printStackTrace(PrintWriter(writer))
+
+        return '\n'.join((super(JavaError, self).__str__(),
+                          "Java stacktrace:", str(writer)))
+
+
 class InvalidArgsError(Exception):
     pass
 
-_jcc._setExceptionTypes(JavaError, InvalidArgsError)
+
+_jcc._set_exception_types(JavaError, InvalidArgsError)
+from _jcc import findClass as _findClass
 from _jcc import *
 
 
-INDENT = '    '
-HALF_INDENT = '  '
+def findClass(className):
+
+    try:
+        cls = _findClass(className)
+    except:
+        print >>sys.stderr, "While loading", className
+        raise
+
+    if cls is None:
+        raise ValueError, className
+
+    return cls
+
+
+INDENT = '  '
+HALF_INDENT = ' '
 
 PRIMITIVES = { 'boolean': 'Z',
                'byte': 'B',
@@ -71,18 +94,39 @@ PRIMITIVES = { 'boolean': 'Z',
                'short': 'S',
                'void': 'V' }
 
-RESERVED = set(['delete', 'and', 'or', 'not', 'xor', 'union', 'NULL',
-                'register', 'const'])
+RESERVED = set(['delete', 'and', 'or', 'not', 'xor', 'union', 'register',
+                'const', 'bool', 'operator', 'typeof', 'asm', 'mutable',
+                'NULL', 'DOMAIN', 'IGNORE'])
+
+RENAME_METHOD_SUFFIX = '_'
+RENAME_FIELD_SUFFIX = '__'
+
+def cppname(name):
+
+    if name in RESERVED:
+        return name + '$'
+
+    return name
+
+
+def cppnames(names):
+
+    return [cppname(name) for name in names]
+
+
+def absname(names):
+
+    if names:
+        return "::%s" %('::'.join(names))
+
+    return ''
 
 
 def typename(cls, current, const):
 
     if cls.isArray():
         componentType = cls.getComponentType()
-        if componentType.isArray():
-            name = 'JArray< %s >' %(typename(componentType, current, False))
-        else:
-            name = 'JArray<%s>' %(typename(componentType, current, False))
+        name = 'JArray< %s >' %(typename(componentType, current, False))
 
     elif cls.isPrimitive():
         name = cls.getName()
@@ -91,13 +135,13 @@ def typename(cls, current, const):
         const = False
 
     elif cls == current:
-        name = cls.getName().split('.')[-1]
+        name = cppname(cls.getName().split('.')[-1])
 
     else:
-        name = cls.getName().replace('.', '::')
+        name = absname([cppname(name) for name in cls.getName().split('.')])
 
     if const:
-        return "const %s&" %(name)
+        return "const %s &" %(name)
 
     return name
 
@@ -123,12 +167,45 @@ def line(out, indent=0, string='', *args):
     out.write('\n')
 
 
-def known(cls, typeset, declares, packages, excludes):
+def known(cls, typeset, declares, packages, excludes, generics):
+
+    if generics:
+        if Class.instance_(cls):
+            cls = Class.cast_(cls)
+        elif ParameterizedType.instance_(cls):
+            pt = ParameterizedType.cast_(cls)
+            if not known(pt.getRawType(), typeset, declares, packages, excludes,
+                         True):
+                return False
+            for ta in pt.getActualTypeArguments():
+                if TypeVariable.instance_(ta):
+                    continue
+                if not known(ta, typeset, declares, packages, excludes, True):
+                    return False
+            return True
+        elif WildcardType.instance_(cls):
+            wc = WildcardType.cast_(cls)
+            for ub in wc.getUpperBounds():
+                if not known(ub, typeset, declares, packages, excludes, True):
+                    return False
+            return True
+        elif TypeVariable.instance_(cls):
+            for bounds in TypeVariable.cast_(cls).getBounds():
+                if not known(bounds, typeset, declares, packages, excludes,
+                             True):
+                    return False
+            return True
+        elif GenericArrayType.instance_(cls):
+            return known(GenericArrayType.cast_(cls).getGenericComponentType(),
+                         typeset, declares, packages, excludes, True)
+        else:
+            raise TypeError, (cls, cls.getClass())
 
     while cls.isArray():
         cls = cls.getComponentType()
 
-    if cls in excludes:
+    className = cls.getName()
+    if className.split('$', 1)[0] in excludes or className in excludes:
         return False
 
     if cls.isPrimitive():
@@ -138,7 +215,7 @@ def known(cls, typeset, declares, packages, excludes):
         declares.add(cls)
         return True
 
-    if split_pkg(cls.getName(), '.')[0] in packages:
+    if split_pkg(className, '.')[0] in packages:
         typeset.add(cls)
         declares.add(cls)
         cls = cls.getSuperclass()
@@ -150,6 +227,37 @@ def known(cls, typeset, declares, packages, excludes):
     return False
 
 
+def addRequiredTypes(cls, typeset, generics):
+
+    if generics:
+        if Class.instance_(cls):
+            cls = Class.cast_(cls)
+            if not (cls.isPrimitive() or cls in typeset):
+                if cls.isArray():
+                    addRequiredTypes(cls.getComponentType(), typeset, True)
+                else:
+                    typeset.add(cls)
+                    cls = cls.getGenericSuperclass()
+                    if cls is not None:
+                        addRequiredTypes(cls, typeset, True)
+        elif ParameterizedType.instance_(cls):
+            pt = ParameterizedType.cast_(cls)
+            addRequiredTypes(pt.getRawType(), typeset, True)
+            for ta in pt.getActualTypeArguments():
+                addRequiredTypes(ta, typeset, True)
+        elif GenericArrayType.instance_(cls):
+            gat = GenericArrayType.cast_(cls)
+            addRequiredTypes(gat.getGenericComponentType(), typeset, True)
+        elif not (TypeVariable.instance_(cls) or WildcardType.instance_(cls)):
+            raise NotImplementedError, repr(cls)
+    else:
+        if cls not in typeset:
+            typeset.add(cls)
+            cls = cls.getSuperclass()
+            if cls is not None:
+                addRequiredTypes(cls, typeset, False)
+
+
 def find_method(cls, name, params):
 
     declared = False
@@ -211,9 +319,9 @@ def forward(out, namespace, indent):
 
     for name, entries in namespace.iteritems():
         if entries is True:
-            line(out, indent, 'class %s;', name)
+            line(out, indent, 'class %s;', cppname(name))
         else:
-            line(out, indent, 'namespace %s {', name)
+            line(out, indent, 'namespace %s {', cppname(name))
             forward(out, entries, indent + 1)
             line(out, indent, '}')
 
@@ -233,30 +341,45 @@ def expandjar(path):
 def jcc(args):
 
     classNames = set()
+    listedClassNames = set()
     packages = set()
     jars = []
-    classpath = []
+    classpath = [_jcc.CLASSPATH]
+    libpath = []
+    vmargs = ['-Djava.awt.headless=true']
     moduleName = None
     modules = []
     build = False
     install = False
     recompile = False
+    egg_info = False
     output = 'build'
     debug = False
     excludes = []
     version = ''
     mappings = {}
     sequences = {}
+    renames = {}
+    use_full_names = False
     env = None
     wrapperFiles = 1
     prefix = None
     root = None
     install_dir = None
+    home_dir = None
     use_distutils = False
     shared = False
     dist = False
+    wininst = False
+    find_jvm_dll = False
     compiler = None
-
+    generics = hasattr(_jcc, "Type")
+    arch = []
+    resources = []
+    imports = {}
+    extra_setup_args = []
+    initvm_args = {}
+    
     i = 1
     while i < len(args):
         arg = args[i]
@@ -276,6 +399,15 @@ def jcc(args):
             elif arg == '--classpath':
                 i += 1
                 classpath.append(args[i])
+            elif arg == '--libpath':
+                i += 1
+                libpath.append(args[i])
+            elif arg == '--vmarg':
+                i += 1
+                vmargs.append(args[i])
+            elif arg == '--maxheap':
+                i += 1
+                initvm_args['maxheap'] = args[i]
             elif arg == '--python':
                 from python import python, module
                 i += 1
@@ -292,6 +424,12 @@ def jcc(args):
             elif arg == '--compile':
                 from python import compile
                 recompile = True
+            elif arg == '--egg-info':
+                from python import compile
+                egg_info = True
+            elif arg == '--extra-setup-arg':
+                i += 1
+                extra_setup_args.append(args[i])
             elif arg == '--output':
                 i += 1
                 output = args[i]
@@ -309,6 +447,12 @@ def jcc(args):
             elif arg == '--sequence':
                 sequences[args[i + 1]] = (args[i + 2], args[i + 3])
                 i += 3
+            elif arg == '--rename':
+                i += 1
+                renames.update(dict([arg.split('=')
+                                     for arg in args[i].split(',')]))
+            elif arg == '--use_full_names':
+                use_full_names = True
             elif arg == '--files':
                 i += 1
                 wrapperFiles = args[i]
@@ -323,6 +467,9 @@ def jcc(args):
             elif arg == '--install-dir':
                 i += 1
                 install_dir = args[i]
+            elif arg == '--home':
+                i += 1
+                home_dir = args[i]
             elif arg == '--use-distutils':
                 use_distutils = True
             elif arg == '--shared':
@@ -330,40 +477,79 @@ def jcc(args):
             elif arg == '--bdist':
                 from python import compile
                 dist = True
+            elif arg == '--wininst':
+                from python import compile
+                wininst = True
+                dist = True
             elif arg == '--compiler':
                 i += 1
                 compiler = args[i]
+            elif arg == '--reserved':
+                i += 1
+                RESERVED.update(args[i].split(','))
+            elif arg == '--arch':
+                i += 1
+                arch.append(args[i])
+            elif arg == '--no-generics':
+                generics = False
+            elif arg == '--find-jvm-dll':
+                find_jvm_dll = True
+            elif arg == '--resources':
+                i += 1
+                resources.append(args[i])
+            elif arg == '--import':
+                i += 1
+                imports[args[i]] = ()
             else:
                 raise ValueError, "Invalid argument: %s" %(arg)
         else:
             classNames.add(arg)
+            listedClassNames.add(arg)
         i += 1
 
-    env = initVM(os.pathsep.join(classpath) or None,
-                 vmargs='-Djava.awt.headless=true')
+    if libpath:
+        vmargs.append('-Djava.library.path=' + os.pathsep.join(libpath))
+
+    initvm_args['maxstack'] = '512k'
+    initvm_args['vmargs'] = vmargs
+
+    env = initVM(os.pathsep.join(classpath) or None, **initvm_args)
 
     typeset = set()
-    excludes = set([findClass(className.replace('.', '/'))
-                    for className in excludes])
+    excludes = set(excludes)
 
-    if recompile or not build and (install or dist):
-        compile(env, os.path.dirname(args[0]), output, moduleName,
-                install, dist, debug, jars, version,
-                prefix, root, install_dir, use_distutils,
-                shared, compiler, modules)
+    if imports:
+        if shared:
+            imports = dict((__import__(import_), set()) for import_ in imports)
+        else:
+            raise ValueError, "--shared must be used when using --import"
+
+    if recompile or not build and (install or dist or egg_info):
+        if moduleName is None:
+            raise ValueError, 'module name not specified (use --python)'
+        else:
+            compile(env, os.path.dirname(args[0]), output, moduleName,
+                    install, dist, debug, jars, version,
+                    prefix, root, install_dir, home_dir, use_distutils,
+                    shared, compiler, modules, wininst, find_jvm_dll,
+                    arch, generics, resources, imports, egg_info,
+                    extra_setup_args)
     else:
-        for className in classNames:
-            cls = findClass(className.replace('.', '/'))
-            if cls is None:
-                raise ValueError, className
-            if cls in excludes:
-                continue
-            if Modifier.isPublic(cls.getModifiers()):
-                typeset.add(cls)
-                cls = cls.getSuperclass()
-                while cls and cls not in typeset:
-                    typeset.add(cls)
-                    cls = cls.getSuperclass()
+        if imports:
+            def walk((include, importset), dirname, names):
+                for name in names:
+                    if name.endswith('.h'):
+                        className = os.path.join(dirname[len(include) + 1:],
+                                                 name[:-2])
+                        if os.path.sep != '/':
+                            className = className.replace(os.path.sep, '/')
+                        importset.add(findClass(className))
+            for import_, importset in imports.iteritems():
+                env._addClassPath(import_.CLASSPATH)
+                include = os.path.join(import_.__dir__, 'include')
+                os.path.walk(include, walk, (include, importset))
+                typeset.update(importset)
+        typeset.add(findClass('java/lang/Object'))
         typeset.add(findClass('java/lang/Class'))
         typeset.add(findClass('java/lang/String'))
         typeset.add(findClass('java/lang/Throwable'))
@@ -372,14 +558,33 @@ def jcc(args):
         if moduleName:
             typeset.add(findClass('java/lang/Number'))
             typeset.add(findClass('java/lang/Boolean'))
+            typeset.add(findClass('java/lang/Byte'))
+            typeset.add(findClass('java/lang/Character'))
+            typeset.add(findClass('java/lang/Double'))
+            typeset.add(findClass('java/lang/Float'))
             typeset.add(findClass('java/lang/Integer'))
             typeset.add(findClass('java/lang/Long'))
-            typeset.add(findClass('java/lang/Double'))
+            typeset.add(findClass('java/lang/Short'))
             typeset.add(findClass('java/util/Iterator'))
             typeset.add(findClass('java/util/Enumeration'))
+            typeset.add(findClass('java/io/StringWriter'))
+            typeset.add(findClass('java/io/PrintWriter'))
+            typeset.add(findClass('java/io/Writer'))
+            packages.add('java.lang')
 
+        for className in classNames:
+            if className.split('$', 1)[0] in excludes or className in excludes:
+                continue
+            cls = findClass(className.replace('.', '/'))
+            if (Modifier.isPublic(cls.getModifiers()) or
+                className in listedClassNames):
+                addRequiredTypes(cls, typeset, generics)
+
+        _dll_export = ''
         if moduleName:
             cppdir = os.path.join(output, '_%s' %(moduleName))
+            if shared and sys.platform == 'win32':
+                _dll_export = "_dll_%s " %(moduleName)
         else:
             cppdir = output
 
@@ -395,8 +600,20 @@ def jcc(args):
                 out_cpp = file(os.path.join(cppdir, fileName), 'w')
 
         done = set()
+        pythonNames = {}
+        for importset in imports.itervalues():
+            done.update(importset)
+            if moduleName:
+                for cls in importset:
+                    name = split_pkg(cls.getName(), '.')[-1]
+                    if not use_full_names:
+                        if name in pythonNames:
+                            raise ValueError, (cls, 'python class name already in use, use --rename', name, pythonNames[name])
+                        else:
+                            pythonNames[name] = cls
+
         todo = typeset - done
-	if allInOne and wrapperFiles > 1:
+        if allInOne and wrapperFiles > 1:
             classesPerFile = max(1, len(todo) / wrapperFiles)
         classCount = 0
         while todo:
@@ -414,23 +631,32 @@ def jcc(args):
                 line(out_h, 0, '#define %s_H', '_'.join(names))
 
                 (superCls, constructors, methods, protectedMethods,
-                 fields, instanceFields, declares) = \
-                    header(env, out_h, cls, typeset, packages, excludes)
+                 methodNames, fields, instanceFields, declares) = \
+                    header(env, out_h, cls, typeset, packages, excludes,
+                           generics, _dll_export)
 
                 if not allInOne:
                     out_cpp = file(fileName + '.cpp', 'w')
                 names, superNames = code(env, out_cpp,
                                          cls, superCls, constructors,
                                          methods, protectedMethods,
-                                         fields, instanceFields, 
-                                         declares, typeset, excludes)
+                                         methodNames, fields, instanceFields, 
+                                         declares, typeset)
                 if moduleName:
+                    name = renames.get(className) or names[-1]
+                    if not use_full_names:
+                        if name in pythonNames:
+                            raise ValueError, (cls, 'python class name already in use, use --rename', name, pythonNames[name])
+                        else:
+                            pythonNames[name] = cls
                     python(env, out_h, out_cpp,
                            cls, superCls, names, superNames,
                            constructors, methods, protectedMethods,
-                           fields, instanceFields,
+                           methodNames, fields, instanceFields,
                            mappings.get(className), sequences.get(className),
-                           declares, typeset, excludes, moduleName)
+                           renames.get(className),
+                           declares, typeset, moduleName, generics,
+                           _dll_export)
 
                 line(out_h)
                 line(out_h, 0, '#endif')
@@ -441,9 +667,9 @@ def jcc(args):
                 elif wrapperFiles > 1:
                     if classCount >= classesPerFile:
                         out_cpp.close()
-	                fileCount += 1
-	                fileName = '__wrap%02d__.cpp' %(fileCount)
-	                out_cpp = file(os.path.join(cppdir, fileName), 'w')
+                        fileCount += 1
+                        fileName = '__wrap%02d__.cpp' %(fileCount)
+                        out_cpp = file(os.path.join(cppdir, fileName), 'w')
                         classCount = 0
                         
             done.update(todo)
@@ -454,27 +680,48 @@ def jcc(args):
 
         if moduleName:
             out = file(os.path.join(cppdir, moduleName) + '.cpp', 'w')
-            module(out, allInOne, done, cppdir, moduleName, shared)
+            module(out, allInOne, done, imports, cppdir, moduleName,
+                   shared, generics, use_full_names)
             out.close()
-            if build or install or dist:
+            if build or install or dist or egg_info:
                 compile(env, os.path.dirname(args[0]), output, moduleName,
                         install, dist, debug, jars, version,
-                        prefix, root, install_dir, use_distutils,
-                        shared, compiler, modules)
+                        prefix, root, install_dir, home_dir, use_distutils,
+                        shared, compiler, modules, wininst, find_jvm_dll,
+                        arch, generics, resources, imports, egg_info,
+                        extra_setup_args)
 
 
-def header(env, out, cls, typeset, packages, excludes):
+def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
 
     names = cls.getName().split('.')
     superCls = cls.getSuperclass()
     declares = set([cls.getClass()])
 
     interfaces = []
-    for interface in cls.getInterfaces():
-        if superCls and interface.isAssignableFrom(superCls):
-            continue
-        if known(interface, typeset, declares, packages, excludes):
-            interfaces.append(interface)
+    if generics:
+        for interface in cls.getGenericInterfaces():
+            if Class.instance_(interface):
+                pt = None
+                interface = Class.cast_(interface)
+            elif ParameterizedType.instance_(interface):
+                pt = ParameterizedType.cast_(interface)
+                interface = Class.cast_(pt.getRawType())
+            else:
+                raise NotImplementedError, repr(interface)
+            if superCls and interface.isAssignableFrom(superCls):
+                continue
+            if known(interface, typeset, declares, packages, excludes, False):
+                interfaces.append(interface)
+                if pt is not None:
+                    for ta in pt.getActualTypeArguments():
+                        addRequiredTypes(ta, typeset, True)
+    else:
+        for interface in cls.getInterfaces():
+            if superCls and interface.isAssignableFrom(superCls):
+                continue
+            if known(interface, typeset, declares, packages, excludes, False):
+                interfaces.append(interface)
 
     if cls.isInterface():
         if interfaces:
@@ -490,11 +737,30 @@ def header(env, out, cls, typeset, packages, excludes):
     constructors = []
     for constructor in cls.getDeclaredConstructors():
         if Modifier.isPublic(constructor.getModifiers()):
-            params = constructor.getParameterTypes()
-            if len(params) == 1 and params[0] == cls:
-                continue
+            if generics:
+                genericParams = constructor.getGenericParameterTypes()
+                params = constructor.getParameterTypes()
+                # It appears that the implicit instance-of-the-declaring-class
+                # parameter of a non-static inner class is missing from 
+                # getGenericParameterTypes()
+                if len(params) == len(genericParams) + 1:
+                    params[1:] = genericParams
+                else:
+                    params = genericParams
+                if len(params) == 1:
+                    if params[0] == cls:
+                        continue
+                    if ParameterizedType.instance_(params[0]):
+                        param = ParameterizedType.cast_(params[0])
+                        if param.getRawType() == cls:
+                            continue
+            else:
+                params = constructor.getParameterTypes()
+                if len(params) == 1 and params[0] == cls:
+                    continue
             for param in params:
-                if not known(param, typeset, declares, packages, excludes):
+                if not known(param, typeset, declares, packages, excludes,
+                             generics):
                     break
             else:
                 constructors.append(constructor)
@@ -505,31 +771,28 @@ def header(env, out, cls, typeset, packages, excludes):
     for method in cls.getDeclaredMethods():
         modifiers = method.getModifiers()
         if Modifier.isPublic(modifiers):
-            returnType = method.getReturnType()
-            if not known(returnType, typeset, declares, packages, excludes):
+            if generics:
+                returnType = method.getGenericReturnType()
+            else:
+                returnType = method.getReturnType()
+            if not known(returnType, typeset, declares, packages, excludes,
+                         generics):
                 continue
             sig = "%s:%s" %(method.getName(), signature(method, True))
             if sig in methods and returnType != cls:
                 continue
-            for param in method.getParameterTypes():
-                if not known(param, typeset, declares, packages, excludes):
+            if generics:
+                params = method.getGenericParameterTypes()
+            else:
+                params = method.getParameterTypes()
+            for param in params:
+                if not known(param, typeset, declares, packages, excludes,
+                             generics):
                     break
             else:
                 methods[sig] = method
         elif Modifier.isProtected(modifiers):
             protectedMethods.append(method)
-    for interface in interfaces:
-        for method in interface.getMethods():
-            sig = "%s:%s" %(method.getName(), signature(method, True))
-            if sig not in methods:
-                param = method.getReturnType()
-                if not known(param, typeset, declares, packages, excludes):
-                    continue
-                for param in method.getParameterTypes():
-                    if not known(param, typeset, declares, packages, excludes):
-                        break
-                else:
-                    methods[sig] = method
 
     def _compare(m0, m1):
         value = cmp(m0.getName(), m1.getName())
@@ -539,21 +802,34 @@ def header(env, out, cls, typeset, packages, excludes):
 
     methods = methods.values()
     sort(methods, fn=_compare)
+    methodNames = set([cppname(method.getName()) for method in methods])
 
     for constructor in constructors:
-        for exception in constructor.getExceptionTypes():
-            known(exception, typeset, declares, packages, excludes)
+        if generics:
+            exceptions = constructor.getGenericExceptionTypes()
+        else:
+            exceptions = constructor.getExceptionTypes()
+        for exception in exceptions:
+            known(exception, typeset, declares, packages, excludes, generics)
     for method in methods:
-        for exception in method.getExceptionTypes():
-            known(exception, typeset, declares, packages, excludes)
+        if generics:
+            exceptions = method.getGenericExceptionTypes()
+        else:
+            exceptions = method.getExceptionTypes()
+        for exception in exceptions:
+            known(exception, typeset, declares, packages, excludes, generics)
 
     fields = []
     instanceFields = []
     for field in cls.getDeclaredFields():
         modifiers = field.getModifiers()
         if Modifier.isPublic(modifiers):
-            if not known(field.getType(),
-                         typeset, declares, packages, excludes):
+            if generics:
+                fieldType = field.getGenericType()
+            else:
+                fieldType = field.getType()
+            if not known(fieldType, typeset, declares, packages, excludes,
+                         generics):
                 continue
             if Modifier.isStatic(modifiers):
                 fields.append(field)
@@ -581,15 +857,16 @@ def header(env, out, cls, typeset, packages, excludes):
     indent = 0;
     line(out)
     for name in names[:-1]:
-        line(out, indent, 'namespace %s {', name)
+        line(out, indent, 'namespace %s {', cppname(name))
         indent += 1
 
     line(out)
     if superClsName == 'JObject':
-        line(out, indent, 'class %s : public JObject {', names[-1])
+        line(out, indent, 'class %s%s : public JObject {',
+             _dll_export, cppname(names[-1]))
     else:
-        line(out, indent, 'class %s : public %s {',
-             names[-1], '::'.join(superNames))
+        line(out, indent, 'class %s%s : public %s {',
+             _dll_export, cppname(names[-1]), absname(cppnames(superNames)))
         
     line(out, indent, 'public:')
     indent += 1
@@ -617,33 +894,31 @@ def header(env, out, cls, typeset, packages, excludes):
         line(out, indent, '};')
 
     line(out)
-    line(out, indent, 'static java::lang::Class *class$;');
+    line(out, indent, 'static ::java::lang::Class *class$;');
     line(out, indent, 'static jmethodID *mids$;');
     if instanceFields:
         line(out, indent, 'static jfieldID *fids$;');
-    line(out, indent, 'static jclass initializeClass();');
-    line(out)
-    line(out, indent, 'static inline java::lang::Class *getclass$() {')
-    line(out, indent + 1, 'initializeClass();')
-    line(out, indent + 1, 'return class$;');
-    line(out, indent, '}')
-
+    line(out, indent, 'static bool live$;');
+    line(out, indent, 'static jclass initializeClass(bool);');
     line(out)
 
     line(out, indent, 'explicit %s(jobject obj) : %s(obj) {',
-         names[-1], '::'.join(superNames))
-    line(out, indent + 1, 'initializeClass();')
+         cppname(names[-1]), absname(cppnames(superNames)))
+    line(out, indent + 1, 'if (obj != NULL)');
+    line(out, indent + 2, 'env->getClass(initializeClass);')
     line(out, indent, '}')
     line(out, indent, '%s(const %s& obj) : %s(obj) {}',
-         names[-1], names[-1], '::'.join(superNames))
+         cppname(names[-1]), cppname(names[-1]),
+         absname(cppnames(superNames)))
 
     if fields:
         line(out)
         for field in fields:
             fieldType = field.getType()
-            fieldName = field.getName()
-            if fieldName in RESERVED:
-                fieldName += '$'
+            fieldName = cppname(field.getName())
+            if fieldName in methodNames:
+                print >>sys.stderr, "  Warning: renaming static variable '%s' on class %s to '%s%s' since it is shadowed by a method of same name." %(fieldName, '.'.join(names), fieldName, RENAME_FIELD_SUFFIX)
+                fieldName += RENAME_FIELD_SUFFIX
             if fieldType.isPrimitive():
                 line(out, indent, 'static %s %s;',
                      typename(fieldType, cls, False), fieldName)
@@ -668,7 +943,7 @@ def header(env, out, cls, typeset, packages, excludes):
         for constructor in constructors:
             params = [typename(param, cls, True)
                       for param in constructor.getParameterTypes()]
-            line(out, indent, '%s(%s);', names[-1], ', '.join(params))
+            line(out, indent, '%s(%s);', cppname(names[-1]), ', '.join(params))
 
     if methods:
         line(out)
@@ -682,9 +957,7 @@ def header(env, out, cls, typeset, packages, excludes):
                 const = ' const'
             params = [typename(param, cls, True)
                       for param in method.getParameterTypes()]
-            methodName = method.getName()
-            if methodName in RESERVED:
-                methodName += '$'
+            methodName = cppname(method.getName())
             line(out, indent, '%s%s %s(%s)%s;',
                  prefix, typename(method.getReturnType(), cls, False),
                  methodName, ', '.join(params), const)
@@ -697,11 +970,11 @@ def header(env, out, cls, typeset, packages, excludes):
         line(out, indent, '}')
 
     return (superCls, constructors, methods, protectedMethods,
-            fields, instanceFields, declares)
+            methodNames, fields, instanceFields, declares)
 
 
 def code(env, out, cls, superCls, constructors, methods, protectedMethods,
-         fields, instanceFields, declares, typeset, excludes):
+         methodNames, fields, instanceFields, declares, typeset):
 
     className = cls.getName()
     names = className.split('.')
@@ -723,34 +996,38 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
     indent = 0
     line(out)
     for name in names[:-1]:
-        line(out, indent, 'namespace %s {', name)
+        line(out, indent, 'namespace %s {', cppname(name))
         indent += 1
 
     line(out)
-    line(out, indent, 'java::lang::Class *%s::class$ = NULL;', names[-1])
-    line(out, indent, 'jmethodID *%s::mids$ = NULL;', names[-1])
+    line(out, indent, '::java::lang::Class *%s::class$ = NULL;',
+         cppname(names[-1]))
+    line(out, indent, 'jmethodID *%s::mids$ = NULL;', cppname(names[-1]))
     if instanceFields:
-        line(out, indent, 'jfieldID *%s::fids$ = NULL;', names[-1])
+        line(out, indent, 'jfieldID *%s::fids$ = NULL;', cppname(names[-1]))
+    line(out, indent, 'bool %s::live$ = false;', cppname(names[-1]))
 
     for field in fields:
         fieldType = field.getType()
-        fieldName = field.getName()
-        if fieldName in RESERVED:
-            fieldName += '$'
+        fieldName = cppname(field.getName())
+        if fieldName in methodNames:
+            fieldName += RENAME_FIELD_SUFFIX
         typeName = typename(fieldType, cls, False)
         if fieldType.isPrimitive():
             line(out, indent, '%s %s::%s = (%s) 0;',
-                 typeName, names[-1], fieldName, typeName)
+                 typeName, cppname(names[-1]), fieldName, typeName)
         else:
             line(out, indent, '%s *%s::%s = NULL;',
-                 typeName, names[-1], fieldName)
+                 typeName, cppname(names[-1]), fieldName)
 
     line(out)
-    line(out, indent, 'jclass %s::initializeClass()', names[-1])
+    line(out, indent, 'jclass %s::initializeClass(bool getOnly)',
+         cppname(names[-1]))
     line(out, indent, '{')
-    line(out, indent + 1, 'if (!class$)')
+    line(out, indent + 1, 'if (getOnly)')
+    line(out, indent + 2, 'return (jclass) (live$ ? class$->this$ : NULL);')
+    line(out, indent + 1, 'if (class$ == NULL)')
     line(out, indent + 1, '{')
-    line(out)
     line(out, indent + 2, 'jclass cls = (jclass) env->findClass("%s");',
          className.replace('.', '/'))
 
@@ -792,7 +1069,7 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
                  fieldName, fieldName, signature(field))
 
     line(out)
-    line(out, indent + 2, 'class$ = (java::lang::Class *) new JObject(cls);')
+    line(out, indent + 2, 'class$ = (::java::lang::Class *) new JObject(cls);')
 
     if fields:
         line(out, indent + 2, 'cls = (jclass) class$->this$;')
@@ -800,18 +1077,21 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
         for field in fields:
             fieldType = field.getType()
             fieldName = field.getName()
+            cppFieldName = cppname(fieldName)
+            if cppFieldName in methodNames:
+                cppFieldName += RENAME_FIELD_SUFFIX
             if fieldType.isPrimitive():
                 line(out, indent + 2,
-                     '%s%s = env->getStatic%sField(cls, "%s");',
-                     fieldName, fieldName in RESERVED and '$' or '',
-                     fieldType.getName().capitalize(), fieldName)
+                     '%s = env->getStatic%sField(cls, "%s");',
+                     cppFieldName, fieldType.getName().capitalize(),
+                     fieldName)
             else:
                 line(out, indent + 2,
-                     '%s%s = new %s(env->getStaticObjectField(cls, "%s", "%s"));',
-                     fieldName, fieldName in RESERVED and '$' or '',
-                     typename(fieldType, cls, False), fieldName,
-                     signature(field))
+                     '%s = new %s(env->getStaticObjectField(cls, "%s", "%s"));',
+                     cppFieldName, typename(fieldType, cls, False),
+                     fieldName, signature(field))
 
+    line(out, indent + 2, "live$ = true;")
     line(out, indent + 1, '}')
     line(out, indent + 1, 'return (jclass) class$->this$;')
     line(out, indent, '}')
@@ -822,7 +1102,8 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
         decls, args = argnames(constructor.getParameterTypes(), cls)
 
         line(out, indent, "%s::%s(%s) : %s(env->newObject(initializeClass, &mids$, mid_init$_%s%s)) {}",
-             names[-1], names[-1], decls, '::'.join(superNames),
+             cppname(names[-1]), cppname(names[-1]), decls,
+             absname(cppnames(superNames)),
              env.strhash(sig), args)
 
     for method in methods:
@@ -848,9 +1129,9 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
             isStatic = False
             if superMethod is not None:
                 qualifier = 'Nonvirtual'
-                this = 'this$, (jclass) %s::class$->this$' %('::'.join(superNames))
+                this = 'this$, (jclass) %s::class$->this$' %(absname(cppnames(superNames)))
                 declaringClass = superMethod.getDeclaringClass()
-                midns = '%s::' %(declaringClass.getName().replace('.', '::'))
+                midns = '%s::' %(typename(declaringClass, cls, False))
             else:
                 qualifier = ''
                 this = 'this$'
@@ -861,12 +1142,13 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
         decls, args = argnames(params, cls)
 
         line(out)
-        line(out, indent, '%s %s::%s%s(%s)%s',
-             typename(returnType, cls, False), names[-1], methodName,
-             methodName in RESERVED and '$' or '', decls, const)
+        line(out, indent, '%s %s::%s(%s)%s',
+             typename(returnType, cls, False), cppname(names[-1]),
+             cppname(methodName), decls, const)
         line(out, indent, '{')
         if isStatic:
-            line(out, indent + 1, 'jclass cls = initializeClass();');
+            line(out, indent + 1,
+                 'jclass cls = env->getClass(initializeClass);');
         if returnType.isPrimitive():
             line(out, indent + 1,
                  '%senv->call%s%sMethod(%s, %smids$[%smid_%s_%s]%s);',
@@ -886,7 +1168,7 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
             fieldName = field.getName()
             line(out)
             line(out, indent, '%s %s::_get_%s() const',
-                 typename(fieldType, cls, False), names[-1], fieldName)
+                 typename(fieldType, cls, False), cppname(names[-1]), fieldName)
             line(out, indent, '{')
             if fieldType.isPrimitive():
                 line(out, indent + 1,
@@ -901,7 +1183,8 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
             if not Modifier.isFinal(field.getModifiers()):
                 line(out)
                 line(out, indent, 'void %s::_set_%s(%s a0) const',
-                     names[-1], fieldName, typename(fieldType, cls, True))
+                     cppname(names[-1]), fieldName,
+                     typename(fieldType, cls, True))
                 line(out, indent, '{')
                 if fieldType.isPrimitive():
                     line(out, indent + 1,
