@@ -1,6 +1,4 @@
 /*
- *   Copyright (c) 2008-2008 Open Source Applications Foundation
- *
  *   Licensed under the Apache License, Version 2.0 (the "License");
  *   you may not use this file except in compliance with the License.
  *   You may obtain a copy of the License at
@@ -27,30 +25,30 @@
 using namespace java::lang;
 
 
-template<class T> class _t_jarray : public t_jarray<T> {
+template<typename T> class _t_JArray : public t_JArray<T> {
 public:
     static PyObject *format;
 };
 
-template<class U>
-static PyObject *get(U *self, int n)
+template<typename U>
+static PyObject *get(U *self, Py_ssize_t n)
 {
     return self->array.get(n);
 }
 
-template<class U>
+template<typename U>
 static PyObject *toSequence(U *self)
 {
     return self->array.toSequence();
 }
 
-template<class U>
-static PyObject *toSequence(U *self, int lo, int hi)
+template<typename U>
+static PyObject *toSequence(U *self, Py_ssize_t lo, Py_ssize_t hi)
 {
     return self->array.toSequence(lo, hi);
 }
 
-template<class U> class _t_iterator {
+template<typename U> class _t_iterator {
 public:
     PyObject_HEAD
     U *obj;
@@ -74,7 +72,7 @@ public:
     static PyTypeObject *JArrayIterator;
 };
 
-template<class T, class U>
+template<typename T, typename U>
 static int init(U *self, PyObject *args, PyObject *kwds)
 {
     PyObject *obj;
@@ -122,14 +120,14 @@ static int init(U *self, PyObject *args, PyObject *kwds)
     return 0;
 }
 
-template<class T, class U>
+template<typename T, typename U>
 static void dealloc(U *self)
 {
     self->array = JArray<T>((jobject) NULL);
     self->ob_type->tp_free((PyObject *) self);
 }
 
-template<class U>
+template<typename U>
 static PyObject *_format(U *self, PyObject *(*fn)(PyObject *))
 {
     if (self->array.this$)
@@ -159,19 +157,19 @@ static PyObject *_format(U *self, PyObject *(*fn)(PyObject *))
     return PyString_FromString("<null>");
 }
 
-template<class U>
+template<typename U>
 static PyObject *repr(U *self)
 {
-    return _format(self, PyObject_Repr);
+    return _format(self, (PyObject *(*)(PyObject *)) PyObject_Repr);
 }
 
-template<class U>
+template<typename U>
 static PyObject *str(U *self)
 {
-    return _format(self, PyObject_Str);
+    return _format(self, (PyObject *(*)(PyObject *)) PyObject_Str);
 }
 
-template<class U>
+template<typename U>
 static int _compare(U *self, PyObject *value, int i0, int i1, int op, int *cmp)
 {
     PyObject *v0 = get<U>(self, i0);
@@ -195,7 +193,7 @@ static int _compare(U *self, PyObject *value, int i0, int i1, int op, int *cmp)
     return 0;
 }
 
-template<class U>
+template<typename U>
 static PyObject *richcompare(U *self, PyObject *value, int op)
 {
     PyObject *result = NULL;
@@ -272,7 +270,7 @@ static PyObject *richcompare(U *self, PyObject *value, int op)
     return result;
 }
 
-template<class U>
+template<typename U>
 static PyObject *iter(U *self)
 {
     _t_iterator<U> *it =
@@ -287,28 +285,28 @@ static PyObject *iter(U *self)
     return (PyObject *) it;
 }
 
-template<class U>
+template<typename U>
 static Py_ssize_t seq_length(U *self)
 {
     if (self->array.this$)
         return self->array.length;
 
-    return -1;
+    return 0;
 }
 
-template<class U>
+template<typename U>
 static PyObject *seq_get(U *self, Py_ssize_t n)
 {
     return get<U>(self, n);
 }
 
-template<class U>
+template<typename U>
 static int seq_contains(U *self, PyObject *value)
 {
     return 0;
 }
 
-template<class U>
+template<typename U>
 static PyObject *seq_concat(U *self, PyObject *arg)
 {
     PyObject *list = toSequence<U>(self);
@@ -323,7 +321,7 @@ static PyObject *seq_concat(U *self, PyObject *arg)
     return list;
 }
 
-template<class U>
+template<typename U>
 static PyObject *seq_repeat(U *self, Py_ssize_t n)
 {
     PyObject *list = toSequence<U>(self);
@@ -338,22 +336,22 @@ static PyObject *seq_repeat(U *self, Py_ssize_t n)
     return list;
 }
 
-template<class U>
+template<typename U>
 static PyObject *seq_getslice(U *self, Py_ssize_t lo, Py_ssize_t hi)
 {
     return toSequence<U>(self, lo, hi);
 }
 
-template<class U>
+template<typename U>
 static int seq_set(U *self, Py_ssize_t n, PyObject *value)
 {
     return self->array.set(n, value);
 }
 
-template<class U>
+template<typename U>
 static int seq_setslice(U *self, Py_ssize_t lo, Py_ssize_t hi, PyObject *values)
 {
-    int length = self->array.length;
+    Py_ssize_t length = self->array.length;
 
     if (values == NULL)
     {
@@ -373,7 +371,7 @@ static int seq_setslice(U *self, Py_ssize_t lo, Py_ssize_t hi, PyObject *values)
     if (!sequence)
         return -1;
 
-    int size = PySequence_Fast_GET_SIZE(sequence);
+    Py_ssize_t size = PySequence_Fast_GET_SIZE(sequence);
     if (size < 0)
         goto error;
 
@@ -383,7 +381,7 @@ static int seq_setslice(U *self, Py_ssize_t lo, Py_ssize_t hi, PyObject *values)
         goto error;
     }
 
-    for (int i = lo; i < hi; i++) {
+    for (Py_ssize_t i = lo; i < hi; i++) {
         PyObject *value = PySequence_Fast_GET_ITEM(sequence, i - lo);
 
         if (value == NULL)
@@ -401,13 +399,13 @@ static int seq_setslice(U *self, Py_ssize_t lo, Py_ssize_t hi, PyObject *values)
     return -1;
 }
 
-template<class T> 
-static jclass initializeClass(void)
+template<typename T> 
+static jclass initializeClass(bool getOnly)
 {
-    return env->get_vm_env()->GetObjectClass(JArray<T>(0).this$);
+    return env->get_vm_env()->GetObjectClass(JArray<T>((Py_ssize_t) 0).this$);
 }
 
-template<class T> 
+template<typename T> 
 static PyObject *cast_(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     PyObject *arg, *clsObj;
@@ -415,7 +413,7 @@ static PyObject *cast_(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if (!PyArg_ParseTuple(args, "O", &arg))
         return NULL;
 
-    if (!PyObject_TypeCheck(arg, &ObjectType))
+    if (!PyObject_TypeCheck(arg, &PY_TYPE(Object)))
     {
         PyErr_SetObject(PyExc_TypeError, arg);
         return NULL;
@@ -444,7 +442,7 @@ static PyObject *cast_(PyTypeObject *type, PyObject *args, PyObject *kwds)
     return JArray<T>(((t_JObject *) arg)->object.this$).wrap();
 }
 
-template<class T> 
+template<typename T> 
 static PyObject *instance_(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     PyObject *arg, *clsObj;
@@ -452,7 +450,7 @@ static PyObject *instance_(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if (!PyArg_ParseTuple(args, "O", &arg))
         return NULL;
 
-    if (!PyObject_TypeCheck(arg, &ObjectType))
+    if (!PyObject_TypeCheck(arg, &PY_TYPE(Object)))
         Py_RETURN_FALSE;
 
     Class argCls = ((t_Object *) arg)->object.getClass();
@@ -472,13 +470,13 @@ static PyObject *instance_(PyTypeObject *type, PyObject *args, PyObject *kwds)
     Py_RETURN_TRUE;
 }
 
-template<class T> 
+template<typename T> 
 static PyObject *assignable_(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     return instance_<T>(type, args, kwds);
 }
 
-template<class T, class U> class jarray_type {
+template< typename T, typename U = _t_JArray<T> > class jarray_type {
 public:
     PySequenceMethods seq_methods;
     PyTypeObject type_object;
@@ -602,38 +600,35 @@ public:
             (richcmpfunc) (PyObject *(*)(U *, PyObject *, int)) richcompare<U>;
         type_object.tp_iter = (getiterfunc) (PyObject *(*)(U *)) iter<U>;
         type_object.tp_methods = methods;
-        type_object.tp_base = &ObjectType;
+        type_object.tp_base = &PY_TYPE(Object);
         type_object.tp_init =
             (initproc) (int (*)(U *, PyObject *, PyObject *)) init<T,U>;
         type_object.tp_new = (newfunc) _new;
     }
 };
 
-template<class T> class _t_jobjectarray : public _t_jarray<T> {
+template<typename T> class _t_jobjectarray : public _t_JArray<T> {
 public:
     PyObject *(*wrapfn)(const T&);
 };
 
-template<>
-static PyObject *get(_t_jobjectarray<jobject> *self, int n)
+template<> PyObject *get(_t_jobjectarray<jobject> *self, Py_ssize_t n)
 {
     return self->array.get(n, self->wrapfn);
 }
 
-template<>
-static PyObject *toSequence(_t_jobjectarray<jobject> *self)
+template<> PyObject *toSequence(_t_jobjectarray<jobject> *self)
 {
     return self->array.toSequence(self->wrapfn);
 }
 
-template<>
-static PyObject *toSequence(_t_jobjectarray<jobject> *self, int lo, int hi)
+template<> PyObject *toSequence(_t_jobjectarray<jobject> *self,
+                                Py_ssize_t lo, Py_ssize_t hi)
 {
     return self->array.toSequence(lo, hi, self->wrapfn);
 }
 
-template<>
-static int init< jobject,_t_jobjectarray<jobject> >(_t_jobjectarray<jobject> *self, PyObject *args, PyObject *kwds)
+template<> int init< jobject,_t_jobjectarray<jobject> >(_t_jobjectarray<jobject> *self, PyObject *args, PyObject *kwds)
 {
     PyObject *obj, *clsObj = NULL;
     PyObject *(*wrapfn)(const jobject &) = NULL;
@@ -644,11 +639,11 @@ static int init< jobject,_t_jobjectarray<jobject> >(_t_jobjectarray<jobject> *se
 
     if (clsObj == NULL)
         cls = env->findClass("java/lang/Object");
-    else if (PyObject_TypeCheck(clsObj, &ClassType))
+    else if (PyObject_TypeCheck(clsObj, &PY_TYPE(Class)))
         cls = (jclass) ((t_Class *) clsObj)->object.this$;
     else if (PyType_Check(clsObj))
     {
-        if (PyType_IsSubtype((PyTypeObject *) clsObj, &JObjectType))
+        if (PyType_IsSubtype((PyTypeObject *) clsObj, &PY_TYPE(JObject)))
         {
             PyObject *cobj = PyObject_GetAttrString(clsObj, "wrapfn_");
 
@@ -722,16 +717,14 @@ static int init< jobject,_t_jobjectarray<jobject> >(_t_jobjectarray<jobject> *se
     return 0;
 }
 
-template<>
-static jclass initializeClass<jobject>(void)
+template<> jclass initializeClass<jobject>(bool getOnly)
 {
     jclass cls = env->findClass("java/lang/Object");
-    return env->get_vm_env()->GetObjectClass(JArray<jobject>(cls, 0).this$);
+    return env->get_vm_env()->GetObjectClass(JArray<jobject>(cls, (Py_ssize_t) 0).this$);
 }
 
-template<> 
-static PyObject *cast_<jobject>(PyTypeObject *type,
-                                PyObject *args, PyObject *kwds)
+template<> PyObject *cast_<jobject>(PyTypeObject *type,
+				    PyObject *args, PyObject *kwds)
 {
     PyObject *arg, *clsArg = NULL;
     PyObject *(*wrapfn)(const jobject&) = NULL;
@@ -740,7 +733,7 @@ static PyObject *cast_<jobject>(PyTypeObject *type,
     if (!PyArg_ParseTuple(args, "O|O", &arg, &clsArg))
         return NULL;
 
-    if (!PyObject_TypeCheck(arg, &ObjectType))
+    if (!PyObject_TypeCheck(arg, &PY_TYPE(Object)))
     {
         PyErr_SetObject(PyExc_TypeError, arg);
         return NULL;
@@ -761,7 +754,7 @@ static PyObject *cast_<jobject>(PyTypeObject *type,
             PyErr_SetObject(PyExc_TypeError, clsArg);
             return NULL;
         }
-        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &JObjectType))
+        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &PY_TYPE(JObject)))
         {
             PyErr_SetObject(PyExc_ValueError, clsArg);
             return NULL;
@@ -800,9 +793,8 @@ static PyObject *cast_<jobject>(PyTypeObject *type,
     return JArray<jobject>(((t_JObject *) arg)->object.this$).wrap(wrapfn);
 }
 
-template<> 
-static PyObject *instance_<jobject>(PyTypeObject *type,
-                                    PyObject *args, PyObject *kwds)
+template<> PyObject *instance_<jobject>(PyTypeObject *type,
+					PyObject *args, PyObject *kwds)
 {
     PyObject *arg, *clsArg = NULL;
     jclass elementCls;
@@ -810,7 +802,7 @@ static PyObject *instance_<jobject>(PyTypeObject *type,
     if (!PyArg_ParseTuple(args, "O|O", &arg, &clsArg))
         return NULL;
 
-    if (!PyObject_TypeCheck(arg, &ObjectType))
+    if (!PyObject_TypeCheck(arg, &PY_TYPE(Object)))
         Py_RETURN_FALSE;
 
     Class argCls = ((t_Object *) arg)->object.getClass();
@@ -825,7 +817,7 @@ static PyObject *instance_<jobject>(PyTypeObject *type,
             PyErr_SetObject(PyExc_TypeError, clsArg);
             return NULL;
         }
-        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &JObjectType))
+        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &PY_TYPE(JObject)))
         {
             PyErr_SetObject(PyExc_ValueError, clsArg);
             return NULL;
@@ -851,9 +843,8 @@ static PyObject *instance_<jobject>(PyTypeObject *type,
     Py_RETURN_TRUE;
 }
 
-template<> 
-static PyObject *assignable_<jobject>(PyTypeObject *type,
-                                      PyObject *args, PyObject *kwds)
+template<> PyObject *assignable_<jobject>(PyTypeObject *type,
+					  PyObject *args, PyObject *kwds)
 {
     PyObject *arg, *clsArg = NULL;
     jclass elementCls;
@@ -861,7 +852,7 @@ static PyObject *assignable_<jobject>(PyTypeObject *type,
     if (!PyArg_ParseTuple(args, "O|O", &arg, &clsArg))
         return NULL;
 
-    if (!PyObject_TypeCheck(arg, &ObjectType))
+    if (!PyObject_TypeCheck(arg, &PY_TYPE(Object)))
         Py_RETURN_FALSE;
 
     Class argCls = ((t_Object *) arg)->object.getClass();
@@ -876,7 +867,7 @@ static PyObject *assignable_<jobject>(PyTypeObject *type,
             PyErr_SetObject(PyExc_TypeError, clsArg);
             return NULL;
         }
-        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &JObjectType))
+        else if (!PyType_IsSubtype((PyTypeObject *) clsArg, &PY_TYPE(JObject)))
         {
             PyErr_SetObject(PyExc_ValueError, clsArg);
             return NULL;
@@ -903,137 +894,187 @@ static PyObject *assignable_<jobject>(PyTypeObject *type,
 }
 
 
-template<class T> PyTypeObject *_t_iterator<T>::JArrayIterator;
-template<class T> PyObject *_t_jarray<T>::format;
+template<typename T> PyTypeObject *_t_iterator<T>::JArrayIterator;
+template<typename T> PyObject *_t_JArray<T>::format;
 
 static jarray_type< jobject, _t_jobjectarray<jobject> > jarray_jobject;
 
-static jarray_type< jstring, _t_jarray<jstring> > jarray_jstring;
-static jarray_type< jboolean, _t_jarray<jboolean> > jarray_jboolean;
-static jarray_type< jbyte, _t_jarray<jbyte> > jarray_jbyte;
-static jarray_type< jchar, _t_jarray<jchar> > jarray_jchar;
-static jarray_type< jdouble, _t_jarray<jdouble> > jarray_jdouble;
-static jarray_type< jfloat, _t_jarray<jfloat> > jarray_jfloat;
-static jarray_type< jint, _t_jarray<jint> > jarray_jint;
-static jarray_type< jlong, _t_jarray<jlong> > jarray_jlong;
-static jarray_type< jshort, _t_jarray<jshort> > jarray_jshort;
+static jarray_type<jstring> jarray_jstring;
+static jarray_type<jboolean> jarray_jboolean;
+static jarray_type<jbyte> jarray_jbyte;
+static jarray_type<jchar> jarray_jchar;
+static jarray_type<jdouble> jarray_jdouble;
+static jarray_type<jfloat> jarray_jfloat;
+static jarray_type<jint> jarray_jint;
+static jarray_type<jlong> jarray_jlong;
+static jarray_type<jshort> jarray_jshort;
 
 
 PyObject *JArray<jobject>::wrap(PyObject *(*wrapfn)(const jobject&))
 {
-    _t_jobjectarray<jobject> *obj =
-        PyObject_New(_t_jobjectarray<jobject>, &jarray_jobject.type_object);
+    if (this$ != NULL)
+    {
+        _t_jobjectarray<jobject> *obj =
+            PyObject_New(_t_jobjectarray<jobject>, &jarray_jobject.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jobject>));
-    obj->array = *this;
-    obj->wrapfn = wrapfn;
+        memset(&(obj->array), 0, sizeof(JArray<jobject>));
+        obj->array = *this;
+        obj->wrapfn = wrapfn;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jstring>::wrap()
+PyObject *JArray<jstring>::wrap() const
 {
-    _t_jarray<jstring> *obj =
-        PyObject_New(_t_jarray<jstring>, &jarray_jstring.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jstring> *obj =
+            PyObject_New(_t_JArray<jstring>, &jarray_jstring.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jstring>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jstring>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jboolean>::wrap()
+PyObject *JArray<jboolean>::wrap() const
 {
-    _t_jarray<jboolean> *obj =
-        PyObject_New(_t_jarray<jboolean>, &jarray_jboolean.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jboolean> *obj =
+            PyObject_New(_t_JArray<jboolean>, &jarray_jboolean.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jboolean>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jboolean>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jbyte>::wrap()
+PyObject *JArray<jbyte>::wrap() const
 {
-    _t_jarray<jbyte> *obj =
-        PyObject_New(_t_jarray<jbyte>, &jarray_jbyte.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jbyte> *obj =
+            PyObject_New(_t_JArray<jbyte>, &jarray_jbyte.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jbyte>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jbyte>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jchar>::wrap()
+PyObject *JArray<jchar>::wrap() const
 {
-    _t_jarray<jchar> *obj =
-        PyObject_New(_t_jarray<jchar>, &jarray_jchar.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jchar> *obj =
+            PyObject_New(_t_JArray<jchar>, &jarray_jchar.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jchar>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jchar>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jdouble>::wrap()
+PyObject *JArray<jdouble>::wrap() const
 {
-    _t_jarray<jdouble> *obj =
-        PyObject_New(_t_jarray<jdouble>, &jarray_jdouble.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jdouble> *obj =
+            PyObject_New(_t_JArray<jdouble>, &jarray_jdouble.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jdouble>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jdouble>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jfloat>::wrap()
+PyObject *JArray<jfloat>::wrap() const
 {
-    _t_jarray<jfloat> *obj =
-        PyObject_New(_t_jarray<jfloat>, &jarray_jfloat.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jfloat> *obj =
+            PyObject_New(_t_JArray<jfloat>, &jarray_jfloat.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jfloat>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jfloat>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jint>::wrap()
+PyObject *JArray<jint>::wrap() const
 {
-    _t_jarray<jint> *obj =
-        PyObject_New(_t_jarray<jint>, &jarray_jint.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jint> *obj =
+            PyObject_New(_t_JArray<jint>, &jarray_jint.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jint>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jint>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jlong>::wrap()
+PyObject *JArray<jlong>::wrap() const
 {
-    _t_jarray<jlong> *obj =
-        PyObject_New(_t_jarray<jlong>, &jarray_jlong.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jlong> *obj =
+            PyObject_New(_t_JArray<jlong>, &jarray_jlong.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jlong>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jlong>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
-PyObject *JArray<jshort>::wrap()
+PyObject *JArray<jshort>::wrap() const
 {
-    _t_jarray<jshort> *obj =
-        PyObject_New(_t_jarray<jshort>, &jarray_jshort.type_object);
+    if (this$ != NULL)
+    {
+        _t_JArray<jshort> *obj =
+            PyObject_New(_t_JArray<jshort>, &jarray_jshort.type_object);
 
-    memset(&(obj->array), 0, sizeof(JArray<jshort>));
-    obj->array = *this;
+        memset(&(obj->array), 0, sizeof(JArray<jshort>));
+        obj->array = *this;
 
-    return (PyObject *) obj;
+        return (PyObject *) obj;
+    }
+
+    Py_RETURN_NONE;
 }
 
 PyObject *JArray_Type(PyObject *self, PyObject *arg)
 {
     PyObject *type_name = NULL, *type;
-    char *name = NULL;
+    char const *name = NULL;
 
     if (PyType_Check(arg))
     {
@@ -1063,10 +1104,11 @@ PyObject *JArray_Type(PyObject *self, PyObject *arg)
     if (type_name != NULL)
     {
         name = PyString_AsString(type_name);
-        Py_DECREF(type_name);
-
         if (!name)
+        {
+            Py_DECREF(type_name);
             return NULL;
+        }
     }
 
     if (!strcmp(name, "object"))
@@ -1092,65 +1134,82 @@ PyObject *JArray_Type(PyObject *self, PyObject *arg)
     else
     {
         PyErr_SetObject(PyExc_ValueError, arg);
+        Py_XDECREF(type_name);
+
         return NULL;
     }
 
     Py_INCREF(type);
+    Py_XDECREF(type_name);
+
     return type;
 }
 
-PyTypeObject *JArrayObjectType;
-PyTypeObject *JArrayStringType;
-PyTypeObject *JArrayBoolType;
-PyTypeObject *JArrayByteType;
-PyTypeObject *JArrayCharType;
-PyTypeObject *JArrayDoubleType;
-PyTypeObject *JArrayFloatType;
-PyTypeObject *JArrayIntType;
-PyTypeObject *JArrayLongType;
-PyTypeObject *JArrayShortType;
+static PyObject *t_JArray_jbyte__get_string_(t_JArray<jbyte> *self, void *data)
+{
+    return self->array.to_string_();
+}
+
+static PyGetSetDef t_JArray_jbyte__fields[] = {
+    { "string_", (getter) t_JArray_jbyte__get_string_, NULL, "", NULL },
+    { NULL, NULL, NULL, NULL, NULL }
+};
+
+
+PyTypeObject *PY_TYPE(JArrayObject);
+PyTypeObject *PY_TYPE(JArrayString);
+PyTypeObject *PY_TYPE(JArrayBool);
+PyTypeObject *PY_TYPE(JArrayByte);
+PyTypeObject *PY_TYPE(JArrayChar);
+PyTypeObject *PY_TYPE(JArrayDouble);
+PyTypeObject *PY_TYPE(JArrayFloat);
+PyTypeObject *PY_TYPE(JArrayInt);
+PyTypeObject *PY_TYPE(JArrayLong);
+PyTypeObject *PY_TYPE(JArrayShort);
+
 
 void _install_jarray(PyObject *module)
 {
     jarray_jobject.install("JArray_object", "object",
                             "__JArray_object_iterator", module);
-    JArrayObjectType = &jarray_jobject.type_object;
+    PY_TYPE(JArrayObject) = &jarray_jobject.type_object;
 
     jarray_jstring.install("JArray_string", "string",
                             "__JArray_string_iterator", module);
-    JArrayStringType = &jarray_jstring.type_object;
+    PY_TYPE(JArrayString) = &jarray_jstring.type_object;
 
     jarray_jboolean.install("JArray_bool", "bool",
                             "__JArray_bool_iterator", module);
-    JArrayBoolType = &jarray_jboolean.type_object;
+    PY_TYPE(JArrayBool) = &jarray_jboolean.type_object;
 
+    jarray_jbyte.type_object.tp_getset = t_JArray_jbyte__fields;
     jarray_jbyte.install("JArray_byte", "byte",
                          "__JArray_byte_iterator", module);
-    JArrayByteType = &jarray_jbyte.type_object;
+    PY_TYPE(JArrayByte) = &jarray_jbyte.type_object;
 
     jarray_jchar.install("JArray_char", "char",
                          "__JArray_char_iterator", module);
-    JArrayCharType = &jarray_jchar.type_object;
+    PY_TYPE(JArrayChar) = &jarray_jchar.type_object;
 
     jarray_jdouble.install("JArray_double", "double",
                            "__JArray_double_iterator", module);
-    JArrayDoubleType = &jarray_jdouble.type_object;
+    PY_TYPE(JArrayDouble) = &jarray_jdouble.type_object;
 
     jarray_jfloat.install("JArray_float", "float",
                           "__JArray_float_iterator", module);
-    JArrayFloatType = &jarray_jfloat.type_object;
+    PY_TYPE(JArrayFloat) = &jarray_jfloat.type_object;
 
     jarray_jint.install("JArray_int", "int",
                         "__JArray_int_iterator", module);
-    JArrayIntType = &jarray_jint.type_object;
+    PY_TYPE(JArrayInt) = &jarray_jint.type_object;
 
     jarray_jlong.install("JArray_long", "long",
                          "__JArray_long_iterator", module);
-    JArrayLongType = &jarray_jlong.type_object;
+    PY_TYPE(JArrayLong) = &jarray_jlong.type_object;
 
     jarray_jshort.install("JArray_short", "short",
                           "__JArray_short_iterator", module);
-    JArrayShortType = &jarray_jshort.type_object;
+    PY_TYPE(JArrayShort) = &jarray_jshort.type_object;
 }
 
 #endif /* PYTHON */
