@@ -1,6 +1,4 @@
 /*
- *   Copyright (c) 2007-2008 Open Source Applications Foundation
- *
  *   Licensed under the Apache License, Version 2.0 (the "License");
  *   you may not use this file except in compliance with the License.
  *   You may obtain a copy of the License at
@@ -21,19 +19,23 @@
 #include <Python.h>
 #include "macros.h"
 
-extern jobjectArray fromPySequence(jclass cls, PyObject *sequence);
-extern PyObject *PyErr_SetJavaError(jthrowable throwable);
+#if PY_VERSION_HEX < 0x02050000
+typedef int Py_ssize_t;
+#endif
 
-extern PyTypeObject *JArrayObjectType;
-extern PyTypeObject *JArrayStringType;
-extern PyTypeObject *JArrayBoolType;
-extern PyTypeObject *JArrayByteType;
-extern PyTypeObject *JArrayCharType;
-extern PyTypeObject *JArrayDoubleType;
-extern PyTypeObject *JArrayFloatType;
-extern PyTypeObject *JArrayIntType;
-extern PyTypeObject *JArrayLongType;
-extern PyTypeObject *JArrayShortType;
+extern jobjectArray fromPySequence(jclass cls, PyObject *sequence);
+extern PyObject *PyErr_SetJavaError();
+
+extern PyTypeObject *PY_TYPE(JArrayObject);
+extern PyTypeObject *PY_TYPE(JArrayString);
+extern PyTypeObject *PY_TYPE(JArrayBool);
+extern PyTypeObject *PY_TYPE(JArrayByte);
+extern PyTypeObject *PY_TYPE(JArrayChar);
+extern PyTypeObject *PY_TYPE(JArrayDouble);
+extern PyTypeObject *PY_TYPE(JArrayFloat);
+extern PyTypeObject *PY_TYPE(JArrayInt);
+extern PyTypeObject *PY_TYPE(JArrayLong);
+extern PyTypeObject *PY_TYPE(JArrayShort);
 
 #endif
 
@@ -41,9 +43,9 @@ extern PyTypeObject *JArrayShortType;
 #include "java/lang/Object.h"
 
 
-template<class T> class JArray : public java::lang::Object {
+template<typename T> class JArray : public java::lang::Object {
 public:
-    int length;
+    Py_ssize_t length;
 
     explicit JArray<T>(jobject obj) : java::lang::Object(obj) {
         length = this$ ? env->getArrayLength((jobjectArray) this$) : 0;
@@ -53,7 +55,7 @@ public:
     }
 
 #ifdef PYTHON
-    JArray<T>(PyObject *sequence) : java::lang::Object(fromPySequence(T::initializeClass(), sequence)) {
+    JArray<T>(PyObject *sequence) : java::lang::Object(fromPySequence(env->getClass(T::initializeClass), sequence)) {
         length = this$ ? env->getArrayLength((jobjectArray) this$) : 0;
     }
 
@@ -68,13 +70,13 @@ public:
 
         PyObject *list = PyList_New(length);
 
-        for (int i = 0; i < length; i++)
+        for (Py_ssize_t i = 0; i < length; i++)
             PyList_SET_ITEM(list, i, (*wrapfn)((*this)[i]));
 
         return list;
     }
 
-    PyObject *get(int n, PyObject *(*wrapfn)(const T&))
+    PyObject *get(Py_ssize_t n, PyObject *(*wrapfn)(const T&))
     {
         if (this$ != NULL)
         {
@@ -90,16 +92,16 @@ public:
     }
 #endif
 
-    T operator[](int n) {
+    T operator[](Py_ssize_t n) {
         return T(env->getObjectArrayElement((jobjectArray) this$, n));
     }
 };
 
 template<> class JArray<jobject> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
-    JArray<jobject>(jclass cls, int n) : java::lang::Object(env->get_vm_env()->NewObjectArray(n, cls, NULL)) {
+    JArray<jobject>(jclass cls, Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewObjectArray(n, cls, NULL)) {
         length = env->getArrayLength((jobjectArray) this$);
     }
 
@@ -112,30 +114,8 @@ template<> class JArray<jobject> : public java::lang::Object {
     }
 
 #ifdef PYTHON
-    JArray<jobject>(jclass cls, PyObject *sequence) : java::lang::Object(env->get_vm_env()->NewObjectArray(PySequence_Length(sequence), cls, NULL)) {
-        length = env->getArrayLength((jobjectArray) this$);
-
-        for (int i = 0; i < length; i++) {
-            PyObject *obj = PySequence_GetItem(sequence, i);
-
-            if (obj == NULL)
-                break;
-
-            if (!PyObject_TypeCheck(obj, &JObjectType))
-            {
-                PyErr_SetObject(PyExc_TypeError, obj);
-                break;
-            }
-
-            jobject jobj = ((t_JObject *) obj)->object.this$;
-
-            Py_DECREF(obj);
-            try {
-                env->setObjectArrayElement((jobjectArray) this$, i, jobj);
-            } catch (JCCEnv::exception e) {
-                PyErr_SetJavaError(e.throwable);
-            }
-        }
+    JArray<jobject>(jclass cls, PyObject *sequence) : java::lang::Object(fromPySequence(cls, sequence)) {
+        length = this$ ? env->getArrayLength((jobjectArray) this$) : 0;
     }
 
     PyObject *toSequence(PyObject *(*wrapfn)(const jobject&))
@@ -143,7 +123,8 @@ template<> class JArray<jobject> : public java::lang::Object {
         return toSequence(0, length, wrapfn);
     }
 
-    PyObject *toSequence(int lo, int hi, PyObject *(*wrapfn)(const jobject&))
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi,
+                         PyObject *(*wrapfn)(const jobject&))
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -161,7 +142,7 @@ template<> class JArray<jobject> : public java::lang::Object {
         if (!wrapfn)
             wrapfn = java::lang::t_Object::wrap_jobject;
 
-        for (int i = lo; i < hi; i++) {
+        for (Py_ssize_t i = lo; i < hi; i++) {
             jobject jobj = env->getObjectArrayElement((jobjectArray) this$, i);
             PyObject *obj = (*wrapfn)(jobj);
 
@@ -171,7 +152,7 @@ template<> class JArray<jobject> : public java::lang::Object {
         return list;
     }
 
-    PyObject *get(int n, PyObject *(*wrapfn)(const jobject&))
+    PyObject *get(Py_ssize_t n, PyObject *(*wrapfn)(const jobject&))
     {
         if (this$ != NULL)
         {
@@ -194,7 +175,7 @@ template<> class JArray<jobject> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -203,19 +184,28 @@ template<> class JArray<jobject> : public java::lang::Object {
 
             if (n >= 0 && n < length)
             {
-                if (!PyObject_TypeCheck(obj, &JObjectType))
+                jobject jobj;
+
+                if (PyString_Check(obj) || PyUnicode_Check(obj))
+                    jobj = env->fromPyString(obj);
+                else if (!PyObject_TypeCheck(obj, &PY_TYPE(JObject)))
                 {
                     PyErr_SetObject(PyExc_TypeError, obj);
                     return -1;
                 }
-
-                jobject jobj = ((t_JObject *) obj)->object.this$;
+                else
+                    jobj = ((t_JObject *) obj)->object.this$;
 
                 try {
                     env->setObjectArrayElement((jobjectArray) this$, n, jobj);
-                } catch (JCCEnv::exception e) {
-                    PyErr_SetJavaError(e.throwable);
-                    return -1;
+                } catch (int e) {
+                    switch (e) {
+                      case _EXC_JAVA:
+                        PyErr_SetJavaError();
+                        return -1;
+                      default:
+                        throw;
+                    }
                 }
 
                 return 0;
@@ -229,14 +219,14 @@ template<> class JArray<jobject> : public java::lang::Object {
     PyObject *wrap(PyObject *(*wrapfn)(const jobject&));
 #endif
 
-    jobject operator[](int n) {
+    jobject operator[](Py_ssize_t n) {
         return (jobject) env->getObjectArrayElement((jobjectArray) this$, n);
     }
 };
 
 template<> class JArray<jstring> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     JArray<jstring>(jobject obj) : java::lang::Object(obj) {
         length = this$ ? env->getArrayLength((jobjectArray) this$) : 0;
@@ -246,7 +236,7 @@ template<> class JArray<jstring> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jstring>(int n) : java::lang::Object(env->get_vm_env()->NewObjectArray(n, env->findClass("java/lang/String"), NULL)) {
+    JArray<jstring>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewObjectArray(n, env->findClass("java/lang/String"), NULL)) {
         length = env->getArrayLength((jobjectArray) this$);
     }
 
@@ -254,7 +244,7 @@ template<> class JArray<jstring> : public java::lang::Object {
     JArray<jstring>(PyObject *sequence) : java::lang::Object(env->get_vm_env()->NewObjectArray(PySequence_Length(sequence), env->findClass("java/lang/String"), NULL)) {
         length = env->getArrayLength((jobjectArray) this$);
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (obj == NULL)
@@ -267,6 +257,7 @@ template<> class JArray<jstring> : public java::lang::Object {
                 break;
 
             env->setObjectArrayElement((jobjectArray) this$, i, str);
+            env->get_vm_env()->DeleteLocalRef(str);
         }
     }
 
@@ -275,7 +266,7 @@ template<> class JArray<jstring> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -290,10 +281,10 @@ template<> class JArray<jstring> : public java::lang::Object {
 
         PyObject *list = PyList_New(hi - lo);
 
-        for (int i = lo; i < hi; i++) {
-            jstring value = (jstring)
+        for (Py_ssize_t i = lo; i < hi; i++) {
+            jstring str = (jstring)
                 env->getObjectArrayElement((jobjectArray) this$, i);
-            PyObject *obj = env->fromJString(value);
+            PyObject *obj = env->fromJString(str, 1);
 
             PyList_SET_ITEM(list, i - lo, obj);
         }
@@ -301,7 +292,7 @@ template<> class JArray<jstring> : public java::lang::Object {
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -312,8 +303,9 @@ template<> class JArray<jstring> : public java::lang::Object {
             {
                 jstring str = (jstring)
                     env->getObjectArrayElement((jobjectArray) this$, n);
+                PyObject *obj = env->fromJString(str, 1);
 
-                return env->fromJString(str);
+                return obj;
             }
         }
 
@@ -321,7 +313,7 @@ template<> class JArray<jstring> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -344,17 +336,17 @@ template<> class JArray<jstring> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jstring operator[](int n) {
+    jstring operator[](Py_ssize_t n) {
         return (jstring) env->getObjectArrayElement((jobjectArray) this$, n);
     }
 };
 
 template<> class JArray<jboolean> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -367,7 +359,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
             elts = env->get_vm_env()->GetBooleanArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseBooleanArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseBooleanArrayElements(array, elts, 0);
         }
         operator jboolean *() {
             return elts;
@@ -386,7 +378,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jboolean>(int n) : java::lang::Object(env->get_vm_env()->NewBooleanArray(n)) {
+    JArray<jboolean>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewBooleanArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -396,7 +388,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         arrayElements elts = elements();
         jboolean *buf = (jboolean *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -421,7 +413,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -438,7 +430,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         arrayElements elts = elements();
         jboolean *buf = (jboolean *) elts;
 
-        for (int i = lo; i < hi; i++) {
+        for (Py_ssize_t i = lo; i < hi; i++) {
             jboolean value = buf[i];
             PyObject *obj = value ? Py_True : Py_False;
 
@@ -449,7 +441,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -464,7 +456,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -482,17 +474,17 @@ template<> class JArray<jboolean> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jboolean operator[](int n) {
+    jboolean operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jboolean *elts = (jboolean *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jboolean value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -500,7 +492,7 @@ template<> class JArray<jboolean> : public java::lang::Object {
 
 template<> class JArray<jbyte> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -513,7 +505,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
             elts = env->get_vm_env()->GetByteArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseByteArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseByteArrayElements(array, elts, 0);
         }
         operator jbyte *() {
             return elts;
@@ -532,7 +524,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jbyte>(int n) : java::lang::Object(env->get_vm_env()->NewByteArray(n)) {
+    JArray<jbyte>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewByteArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -545,7 +537,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
         if (PyString_Check(sequence))
             memcpy(buf, PyString_AS_STRING(sequence), length);
         else
-            for (int i = 0; i < length; i++) {
+            for (Py_ssize_t i = 0; i < length; i++) {
                 PyObject *obj = PySequence_GetItem(sequence, i);
 
                 if (!obj)
@@ -556,6 +548,11 @@ template<> class JArray<jbyte> : public java::lang::Object {
                     buf[i] = (jbyte) PyString_AS_STRING(obj)[0];
                     Py_DECREF(obj);
                 }
+                else if (PyInt_CheckExact(obj))
+                {
+                    buf[i] = (jbyte) PyInt_AS_LONG(obj);
+                    Py_DECREF(obj);
+                }
                 else
                 {
                     PyErr_SetObject(PyExc_TypeError, obj);
@@ -575,7 +572,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -590,11 +587,26 @@ template<> class JArray<jbyte> : public java::lang::Object {
 
         arrayElements elts = elements();
         jbyte *buf = (jbyte *) elts;
+        PyObject *tuple = PyTuple_New(hi - lo);
+        
+        for (Py_ssize_t i = 0; i < hi - lo; i++)
+            PyTuple_SET_ITEM(tuple, i, PyInt_FromLong(buf[lo + i]));
 
-        return PyString_FromStringAndSize((char *) buf + lo, hi - lo);
+        return tuple;
     }
 
-    PyObject *get(int n)
+    PyObject *to_string_()
+    {
+        if (this$ == NULL)
+            Py_RETURN_NONE;
+
+        arrayElements elts = elements();
+        jbyte *buf = (jbyte *) elts;
+
+        return PyString_FromStringAndSize((char *) buf, length);
+    }
+
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -604,7 +616,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
             if (n >= 0 && n < length)
             {
                 jbyte b = (*this)[n];
-                return PyString_FromStringAndSize((char *) &b, 1);
+                return PyInt_FromLong(b);
             }
         }
 
@@ -612,7 +624,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -621,18 +633,13 @@ template<> class JArray<jbyte> : public java::lang::Object {
 
             if (n >= 0 && n < length)
             {
-                if (!PyString_Check(obj))
+                if (!PyInt_CheckExact(obj))
                 {
                     PyErr_SetObject(PyExc_TypeError, obj);
                     return -1;
                 }
-                if (PyString_GET_SIZE(obj) != 1)
-                {
-                    PyErr_SetObject(PyExc_ValueError, obj);
-                    return -1;
-                }
 
-                elements()[n] = (jbyte) PyString_AS_STRING(obj)[0];
+                elements()[n] = (jbyte) PyInt_AS_LONG(obj);
                 return 0;
             }
         }
@@ -641,17 +648,17 @@ template<> class JArray<jbyte> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jbyte operator[](int n) {
+    jbyte operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jbyte *elts = (jbyte *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jbyte value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -659,7 +666,7 @@ template<> class JArray<jbyte> : public java::lang::Object {
 
 template<> class JArray<jchar> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -672,7 +679,7 @@ template<> class JArray<jchar> : public java::lang::Object {
             elts = env->get_vm_env()->GetCharArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseCharArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseCharArrayElements(array, elts, 0);
         }
         operator jchar *() {
             return elts;
@@ -691,7 +698,7 @@ template<> class JArray<jchar> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jchar>(int n) : java::lang::Object(env->get_vm_env()->NewCharArray(n)) {
+    JArray<jchar>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewCharArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -709,12 +716,12 @@ template<> class JArray<jchar> : public java::lang::Object {
             else
             {
                 Py_UNICODE *pchars = PyUnicode_AS_UNICODE(sequence);
-                for (int i = 0; i < length; i++)
+                for (Py_ssize_t i = 0; i < length; i++)
                     buf[i] = (jchar) pchars[i];
             }
         }
         else
-            for (int i = 0; i < length; i++) {
+            for (Py_ssize_t i = 0; i < length; i++) {
                 PyObject *obj = PySequence_GetItem(sequence, i);
 
                 if (!obj)
@@ -739,7 +746,7 @@ template<> class JArray<jchar> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -763,14 +770,14 @@ template<> class JArray<jchar> : public java::lang::Object {
             PyObject *string = PyUnicode_FromUnicode(NULL, hi - lo);
             Py_UNICODE *pchars = PyUnicode_AS_UNICODE(string);
 
-            for (int i = lo; i < hi; i++)
+            for (Py_ssize_t i = lo; i < hi; i++)
                 pchars[i - lo] = (Py_UNICODE) buf[i];
 
             return string;
         }
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -799,7 +806,7 @@ template<> class JArray<jchar> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -828,17 +835,17 @@ template<> class JArray<jchar> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jchar operator[](int n) {
+    jchar operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jchar *elts = (jchar *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jchar value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -846,7 +853,7 @@ template<> class JArray<jchar> : public java::lang::Object {
 
 template<> class JArray<jdouble> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -859,7 +866,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
             elts = env->get_vm_env()->GetDoubleArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseDoubleArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseDoubleArrayElements(array, elts, 0);
         }
         operator jdouble *() {
             return elts;
@@ -878,7 +885,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jdouble>(int n) : java::lang::Object(env->get_vm_env()->NewDoubleArray(n)) {
+    JArray<jdouble>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewDoubleArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -888,7 +895,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
         arrayElements elts = elements();
         jdouble *buf = (jdouble *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -913,7 +920,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -930,13 +937,13 @@ template<> class JArray<jdouble> : public java::lang::Object {
         arrayElements elts = elements();
         jdouble *buf = (jdouble *) elts;
 
-        for (int i = lo; i < hi; i++)
+        for (Py_ssize_t i = lo; i < hi; i++)
             PyList_SET_ITEM(list, i - lo, PyFloat_FromDouble((double) buf[i]));
 
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -951,7 +958,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -975,17 +982,17 @@ template<> class JArray<jdouble> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jdouble operator[](int n) {
+    jdouble operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jdouble *elts = (jdouble *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jdouble value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -993,7 +1000,7 @@ template<> class JArray<jdouble> : public java::lang::Object {
 
 template<> class JArray<jfloat> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -1006,7 +1013,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
             elts = env->get_vm_env()->GetFloatArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseFloatArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseFloatArrayElements(array, elts, 0);
         }
         operator jfloat *() {
             return elts;
@@ -1025,7 +1032,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jfloat>(int n) : java::lang::Object(env->get_vm_env()->NewFloatArray(n)) {
+    JArray<jfloat>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewFloatArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -1035,7 +1042,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
         arrayElements elts = elements();
         jfloat *buf = (jfloat *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -1060,7 +1067,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -1077,13 +1084,13 @@ template<> class JArray<jfloat> : public java::lang::Object {
         arrayElements elts = elements();
         jfloat *buf = (jfloat *) elts;
 
-        for (int i = lo; i < hi; i++)
+        for (Py_ssize_t i = lo; i < hi; i++)
             PyList_SET_ITEM(list, i - lo, PyFloat_FromDouble((double) buf[i]));
 
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -1098,7 +1105,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -1122,17 +1129,17 @@ template<> class JArray<jfloat> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jfloat operator[](int n) {
+    jfloat operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jfloat *elts = (jfloat *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jfloat value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -1140,7 +1147,7 @@ template<> class JArray<jfloat> : public java::lang::Object {
 
 template<> class JArray<jint> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -1153,7 +1160,7 @@ template<> class JArray<jint> : public java::lang::Object {
             elts = env->get_vm_env()->GetIntArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseIntArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseIntArrayElements(array, elts, 0);
         }
         operator jint *() {
             return elts;
@@ -1172,7 +1179,7 @@ template<> class JArray<jint> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jint>(int n) : java::lang::Object(env->get_vm_env()->NewIntArray(n)) {
+    JArray<jint>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewIntArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -1182,7 +1189,7 @@ template<> class JArray<jint> : public java::lang::Object {
         arrayElements elts = elements();
         jint *buf = (jint *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -1207,7 +1214,7 @@ template<> class JArray<jint> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -1224,13 +1231,13 @@ template<> class JArray<jint> : public java::lang::Object {
         arrayElements elts = elements();
         jint *buf = (jint *) elts;
 
-        for (int i = lo; i < hi; i++)
+        for (Py_ssize_t i = lo; i < hi; i++)
             PyList_SET_ITEM(list, i - lo, PyInt_FromLong(buf[i]));
 
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -1245,7 +1252,7 @@ template<> class JArray<jint> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -1269,17 +1276,17 @@ template<> class JArray<jint> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jint operator[](int n) {
+    jint operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jint *elts = (jint *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jint value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -1287,7 +1294,7 @@ template<> class JArray<jint> : public java::lang::Object {
 
 template<> class JArray<jlong> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -1300,7 +1307,7 @@ template<> class JArray<jlong> : public java::lang::Object {
             elts = env->get_vm_env()->GetLongArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseLongArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseLongArrayElements(array, elts, 0);
         }
         operator jlong *() {
             return elts;
@@ -1319,7 +1326,7 @@ template<> class JArray<jlong> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jlong>(int n) : java::lang::Object(env->get_vm_env()->NewLongArray(n)) {
+    JArray<jlong>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewLongArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -1329,7 +1336,7 @@ template<> class JArray<jlong> : public java::lang::Object {
         arrayElements elts = elements();
         jlong *buf = (jlong *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -1354,7 +1361,7 @@ template<> class JArray<jlong> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -1371,13 +1378,13 @@ template<> class JArray<jlong> : public java::lang::Object {
         arrayElements elts = elements();
         jlong *buf = (jlong *) elts;
 
-        for (int i = lo; i < hi; i++)
+        for (Py_ssize_t i = lo; i < hi; i++)
             PyList_SET_ITEM(list, i - lo, PyLong_FromLongLong((long long) buf[i]));
 
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -1392,7 +1399,7 @@ template<> class JArray<jlong> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -1416,7 +1423,7 @@ template<> class JArray<jlong> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
     jlong operator[](long n) {
@@ -1426,7 +1433,7 @@ template<> class JArray<jlong> : public java::lang::Object {
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jlong value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -1434,7 +1441,7 @@ template<> class JArray<jlong> : public java::lang::Object {
 
 template<> class JArray<jshort> : public java::lang::Object {
   public:
-    int length;
+    Py_ssize_t length;
 
     class arrayElements {
     private:
@@ -1447,7 +1454,7 @@ template<> class JArray<jshort> : public java::lang::Object {
             elts = env->get_vm_env()->GetShortArrayElements(array, &isCopy);
         }
         virtual ~arrayElements() {
-            env->get_vm_env()->ReleaseShortArrayElements(array, elts, isCopy);
+            env->get_vm_env()->ReleaseShortArrayElements(array, elts, 0);
         }
         operator jshort *() {
             return elts;
@@ -1466,7 +1473,7 @@ template<> class JArray<jshort> : public java::lang::Object {
         length = obj.length;
     }
 
-    JArray<jshort>(int n) : java::lang::Object(env->get_vm_env()->NewShortArray(n)) {
+    JArray<jshort>(Py_ssize_t n) : java::lang::Object(env->get_vm_env()->NewShortArray(n)) {
         length = env->getArrayLength((jarray) this$);
     }
 
@@ -1476,7 +1483,7 @@ template<> class JArray<jshort> : public java::lang::Object {
         arrayElements elts = elements();
         jshort *buf = (jshort *) elts;
 
-        for (int i = 0; i < length; i++) {
+        for (Py_ssize_t i = 0; i < length; i++) {
             PyObject *obj = PySequence_GetItem(sequence, i);
 
             if (!obj)
@@ -1501,7 +1508,7 @@ template<> class JArray<jshort> : public java::lang::Object {
         return toSequence(0, length);
     }
 
-    PyObject *toSequence(int lo, int hi)
+    PyObject *toSequence(Py_ssize_t lo, Py_ssize_t hi)
     {
         if (this$ == NULL)
             Py_RETURN_NONE;
@@ -1518,13 +1525,13 @@ template<> class JArray<jshort> : public java::lang::Object {
         arrayElements elts = elements();
         jshort *buf = (jshort *) elts;
 
-        for (int i = lo; i < hi; i++)
+        for (Py_ssize_t i = lo; i < hi; i++)
             PyList_SET_ITEM(list, i - lo, PyInt_FromLong(buf[i]));
 
         return list;
     }
 
-    PyObject *get(int n)
+    PyObject *get(Py_ssize_t n)
     {
         if (this$ != NULL)
         {
@@ -1539,7 +1546,7 @@ template<> class JArray<jshort> : public java::lang::Object {
         return NULL;
     }
 
-    int set(int n, PyObject *obj)
+    int set(Py_ssize_t n, PyObject *obj)
     {
         if (this$ != NULL)
         {
@@ -1563,17 +1570,17 @@ template<> class JArray<jshort> : public java::lang::Object {
         return -1;
     }
 
-    PyObject *wrap();
+    PyObject *wrap() const;
 #endif
 
-    jshort operator[](int n) {
+    jshort operator[](Py_ssize_t n) {
         JNIEnv *vm_env = env->get_vm_env();
         jboolean isCopy = 0;
         jshort *elts = (jshort *)
             vm_env->GetPrimitiveArrayCritical((jarray) this$, &isCopy);
         jshort value = elts[n];
 
-        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, isCopy);
+        vm_env->ReleasePrimitiveArrayCritical((jarray) this$, elts, 0);
 
         return value;
     }
@@ -1581,10 +1588,18 @@ template<> class JArray<jshort> : public java::lang::Object {
 
 #ifdef PYTHON
 
-template<class T> class t_jarray {
+template<typename T> class t_JArray {
 public:
     PyObject_HEAD
     JArray<T> array;
+
+    static PyObject *wrap_Object(const JArray<T>& array)
+    {
+        if (!!array)
+            return array.wrap();
+
+        Py_RETURN_NONE;
+    }
 };
 
 #endif
